function lab = DisRFC_clust(emb,nclus,param)
% lab = DisRFC_clust(emb,nclus,param);
% Perform clustering in the embedding
% 
% Dissimilarity Random Forest Clustering
% (c) Manuele Bicego 2020
% For all details please refer to
% M. Bicego: "Dissimilarity Random Forest Clustering",
% Proc. Int. Conf. on Data Mining (ICDM2020)
%
%
% Input parameters:
% emb   -> the embedding (obtained with DisRFC_emed.m)
% nclus -> number of clusters
% param -> parameters
% 
% Output:
% lab   -> the assigned labels
 
sig = emb;
mtd = param.ensC;
if param.verbose, fprintf('Clustering \n'), end
switch mtd
    case 1
       % 1: KM-SED: Kmeans with Squared Euclidean Distance
       % 
       % A. Topchy, A.K. Jain, and W. Punch
       % Combining Multiple Weak Clusterings
       % ICDM 2003
       % 1. First Step: normalization of categorical features
       % following:
       % B. Mirkin, “Reinterpreting the Category Utility Function”,
       % Machine Learning, 45(2), 2001, pp. 219-228.
       [nO,nT] = size(sig);
       csig = num2cell(sig,1);
       cndif = cellfun(@unique,csig,'UniformOutput',false);
       mx = cellfun(@length,cndif);
       Nsig = zeros(nO,sum(mx));
       ini = 1;
       eni = mx(1);
       for t = 1:nT
           pi = sig(:,t);
           xu = unique(pi);
           pi2 = zeros(size(pi));
           for j = 1:length(xu)
               pi2(pi==xu(j)) = j;
           end
           a = sub2ind([nO,mx(t)],[1:nO]',pi2);
           b = zeros(nO,max(pi2));
           b(a) = 1;
           Nsig(:,ini:eni) = b;
           if t<nT
               ini = ini+mx(t);
               eni = eni+mx(t+1);
           end
       end
       mn = mean(Nsig);
       Nsig = Nsig - repmat(mn,[nO,1]);
       [T,C,sumd] = kmeans(Nsig, nclus, 'EmptyAction', 'singleton','Replicates',param.nrandKM);
       errf = sum(sumd);
    case 2 % 2: KM-KL:  K-means with KL divergence
        % K-means Based Consensus Clustering
        % Wu et al: "K-means Based Consensus Clustering; a unified view",
        % IEEE Tr. on TKDD, vol 27(1), 2015
        %
        KCCmtd = 1; % KCC-Uc (see tab 3 of the paper)
        [T,errf] = DisRFC_KCC(sig,KCCmtd,nclus,param.nrandKM,param.verbose);
 
    case 3
        % here we consider cluster memberships as categorical features
        % and apply categorical clustering
        [T,C,sumd] = kmedoids(sig,nclus,'Distance','hamming','Replicates',param.nrandKM);
        errf = sum(sumd);
    case 4
        % Kmodes algorithm for categorical data clustering
        % ZHEXUE HUANG, Data Mining and Knowledge Discovery 2, 283--304 (1998)
        [T,errf] = DisRFC_kmodes(sig,nclus,param.nrandKM);
     
end
lab = T;


