function emb = DisRFC_embed(D,Rf,param)
% emb = DisRFC_embed(D,Rf,param)
% Embed the objects using the trained UDRF
%
% Dissimilarity Random Forest Clustering
% (c) Manuele Bicego 2020
% For all details please refer to
% M. Bicego: "Dissimilarity Random Forest Clustering",
% Proc. Int. Conf. on Data Mining (ICDM2020)
%
%
% Input parameters:
% D     -> pairwise distance between points% param -> parameters for the training
%          For standard parameter simply use
%          param = DisRFC_defaultParam;
% RF    -> trained UD-RF (with DisRFC_trainUDRF.m)
% param -> parameters
%
% Output:
% emb   -> the embedding


n = size(D,1);
objs = 1:n;
T = length(Rf.trees);
if param.verbose
    fprintf('Embedding (UDRF with %d trees): ',param.ntrees)
end
switch param.emb
    case 1
        % OP embedding at leave levels
        sig = zeros(n,T);
        for tr = 1:T
            if param.verbose
                if (mod(tr,10) == 0) || (tr == 1) || (tr == param.ntrees),fprintf('.');end;
            end
            for i = 1:n
                o = objs(i);
                v = Rf.trees{tr};
                ndx = 0;
                while (1)
                    ndx = ndx+1;
                    if (isempty(v.l)) % leaf node
                        sig(i,tr) = v.lid;
                        break
                    else
                        if (D(o,v.oL)<D(o,v.oR))
                            v = v.l;
                        else
                            v = v.r;
                        end
                    end
                end
            end
        end
        emb = sig;
        
        
    case 2
        % MP embedding with MPparam levels
        sig5 = cell(T,1);
        MPparam = Rf.param.MPparam;
        for tr = 1:T
            if param.verbose
                if (mod(tr,10) == 0) || (tr == 1) || (tr == param.ntrees),fprintf('.');end;
            end
            sig4 = zeros(n,Rf.trees{tr}.maxdepth-1);
            for i = 1:n
                o = objs(i);
                v = Rf.trees{tr};
                ndx = 0;
                while (1)
                    ndx = ndx+1;
                    if ndx>1 %exclude root (always present)
                        sig4(i,ndx-1) = v.idx;
                    end
                    if (isempty(v.l)) % leaf node
                        break
                    else
                        if (D(o,v.oL)<D(o,v.oR))
                            v = v.l;
                        else
                            v = v.r;
                        end
                    end
                end
                for j=ndx:Rf.trees{tr}.maxdepth-1
                    sig4(i,j) = sig4(i,j-1);
                end
            end
            a = Rf.trees{tr}.maxdepth-1;
            st = a/(MPparam+1);
            indx = round(2*st:st:a);
            sig5{tr} = sig4(:,indx);
        end
        emb = [sig5{:}];
end
if param.verbose
    fprintf('\n');
end