function t = DisRFC_trainUDT(D,NNs,objs,param,idx)
% t = DisRFC_trainUDT(D,NNs,objs,param,idx)
% Train an Unsupervised Dissimilarity Tree
%
% Dissimilarity Random Forest Clustering
% (c) Manuele Bicego 2020
% For all details please refer to
% M. Bicego: "Dissimilarity Random Forest Clustering", 
% Proc. Int. Conf. on Data Mining (ICDM2020)
%
% (routine used inside DisRFC_trainUDRF.m)
% Input parameters:
% D     -> pairwise distance between points% param -> parameters for the training
%          For standard parameter simply use
%          param = DisRFC_defaultParam;
% NNs   -> nearest neighbors
% objs  -> objects used to train the tree
% param -> parameters
% idx   -> identifier of the root node
%
% Output:
% t     -> the trained tree


global lid
meth = param.UDT;
n = length(objs);
t.idx = idx;
stopsplit = 0;
switch meth
    case 1
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % meth = 1: "Rand" scheme
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if length(dec2bin(idx)) >= param.maxDepth
            % max length
            stopsplit = 1;
        elseif length(objs) < param.minsizeSplit
            % leave with #elements < minleafsize
            stopsplit = 1;
        else
            sptot = 0;
            while 1
                sptot = sptot +1;
                if sptot > 500
                    stopsplit = 1;
                    break;
                end
                %random pick of two exemplars
                a = randperm(length(objs));
                oL = objs(a(1));
                oR = objs(a(2));
                
                % splitting
                objsL = objs((D(objs,oL) < D(objs,oR)));
                objsR = setdiff(objs,objsL);
                ntL = length(objsL);
                ntR = length(objsR);
                % check for a proper split
                if ntL>=1 && ntR >=1
                    break
                end
            end
        end
    case 2 
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % meth = 2: "HaussD" scheme 
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        maxTrialsSplit = param.maxTrialsSplit;
        if length(dec2bin(idx)) >= param.maxDepth
            % max depth
            stopsplit = 1;
        elseif length(objs) < param.minsizeSplit
            % leave with #elements < minleafsize (or less than renyidivK)
            stopsplit = 1;
        else
            bestgain = -inf;
            gain = zeros(1,maxTrialsSplit);
            sp = 0;
            sptot = 0;
            while 1
                % generating random pair
                a = randperm(length(objs));
                XoL = objs(a(1));
                XoR = objs(a(2));
                % splitting
                XobjsL = objs((D(objs,XoL) < D(objs,XoR)));
                XobjsR = setdiff(objs,XobjsL);
                ntL = length(XobjsL);
                ntR = length(XobjsR);
                if ntL>=1 && ntR>=1
                    sp = sp+1;
                    % computing symmetric Haussdorf Distance
                    hd = D(XobjsL,XobjsR);
                    haus1 = max(min(hd,[],2));
                    hd = D(XobjsR,XobjsL);
                    haus2 = max(min(hd,[],2));
                    gain(sp) = (haus1+haus2)/2;
                    if gain(sp) > bestgain
                        bestgain = gain(sp);
                        oL = XoL;
                        oR = XoR;
                    end
                    if sp == maxTrialsSplit
                        break;
                    end
                else
                    sptot = sptot+1;
                    if sptot > 500
                        if sp == 0
                            % no splits found
                            % random split
                            while 1
                                 sptot = sptot+1;
                                 a = randperm(length(objs));
                                 oL = objs(a(1));
                                 oR = objs(a(2));
                                 objsL = objs((D(objs,oL) < D(objs,oR)));
                                 objsR = setdiff(objs,objsL);
                                 ntL = length(objsL);
                                 ntR = length(objsR);
                                 % check for a proper split
                                 if ntL>=1 && ntR >=1
                                     break
                                 end
                             end 
                        end
                        break;
                    end
                end
            end
            % splitting
            objsL = objs((D(objs,oL) < D(objs,oR)));
            objsR = setdiff(objs,objsL);
        end
    case 3
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % meth = 3: "RenyiD" scheme
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        maxTrialsSplit = param.maxTrialsSplit;
        renyidivAlpha = param.renyidivAlpha;
        if param.renyidivK == -1
            renyidivK = round(sqrt(n));
        else
            renyidivK = param.renyidivK;
        end
        if length(dec2bin(idx)) >= param.maxDepth
            % max depth
            stopsplit = 1;
        elseif length(objs) < max(param.minsizeSplit,(2*(renyidivK+1)+5))
            % leave with #elements < minleafsize (or less than a reasonable value 
            % depending on renyidivK)
            stopsplit = 1;
        else
            bestgain = -inf;
            gain = zeros(1,maxTrialsSplit);
            sp = 0;
            sptot = 0;
            sneg = 0;
            while 1
                % generating random pair
                a = randperm(length(objs));
                XoL = objs(a(1));
                XoR = objs(a(2));
                % splitting
                XobjsL = objs((D(objs,XoL) < D(objs,XoR)));
                XobjsR = setdiff(objs,XobjsL);
                % computing entropy gain
                ntL = length(XobjsL);
                ntR = length(XobjsR);
                if ntL>renyidivK && ntR>renyidivK
                    sp = sp+1;
                    % computing divergence
                    for j = 1:2
                        if j == 1
                            X = XobjsL;
                            Y = XobjsR;
                        else
                            X = XobjsR;
                            Y = XobjsL;
                        end
                        M = length(Y);
                        N = length(X);
                        eta = M/N;
                        labels= zeros(size(D,1),1);
                        labels(Y) = 1;
                        labels(X) = 2;
                        xNNs = NNs;
                        xNNs(:,labels==0) = [];
                        Ne = xNNs(Y,1:renyidivK); % distance Y from all neighbors
                        lNe = labels(Ne);
                        Mi = sum(lNe==1,2);
                        Ni = sum(lNe==2,2);
                        a1 = eps+(eta^renyidivAlpha)/M*sum(((Ni./(Mi+1)).^renyidivAlpha));
                        Xgain(j) = 1/(renyidivAlpha-1)*log(a1);
                    end
                    gain(sp) = mean(Xgain);
                    if gain(sp) > bestgain
                        bestgain = gain(sp);
                        oL = XoL;
                        oR = XoR;
                    end
                    if sp == maxTrialsSplit
                        break;
                    end
                end
                sptot = sptot+1;
                if sptot > maxTrialsSplit*2
                    if sp == 0
                        % no splits found, random split
                        sptot2 = 0;
                        while 1
                            sptot2 = sptot2 +1;
                            if sptot2 > 500
                                stopsplit = 1;
                                break;
                            end
                            %random pick of two exemplars
                            a = randperm(length(objs));
                            oL = objs(a(1));
                            oR = objs(a(2));
                            
                            % splitting
                            objsL = objs((D(objs,oL) < D(objs,oR)));
                            objsR = setdiff(objs,objsL);
                            ntL = length(objsL);
                            ntR = length(objsR);
                            % check for a proper split
                            if ntL>=1 && ntR >=1
                                break
                            end
                        end
                    end
                    break;
                end
            end
            % fprintf('%d %d %d\n',sp,sneg,sptot);
            % splitting
            objsL = objs((D(objs,oL) < D(objs,oR)));
            objsR = setdiff(objs,objsL);
        end
end
if stopsplit
    % leave
    t.l = [];
    t.r = [];
    t.lid = lid;
    lid = lid + 1;
    t.sizet = length(objs);
    t.nleaves = 1;
    t.maxdepth = ceil(log2(idx));
else
    t.oL = oL;
    t.oR = oR;
    t.l = DisRFC_trainUDT(D,NNs,objsL,param,idx*2);
    t.r = DisRFC_trainUDT(D,NNs,objsR,param,idx*2+1);
    t.sizet = size(objs,1);
    t.nleaves = t.l.nleaves + t.r.nleaves;
    t.maxdepth = max(t.l.maxdepth,t.r.maxdepth);
end



